@@ -44,7 +44,7 @@ Manipulating module defaults:
     MIME::Charset::default("iso-8859-1");
     MIME::Charset::fallback("us-ascii");
 
-Non-OOP functions (may be deprecated in near future):
+Non-OO functions (may be deprecated in near future):
 
     use MIME::Charset qw(:info);
 
@@ -77,7 +77,7 @@ Non-OOP functions (may be deprecated in near future):
 MIME::Charset provides informations about character sets used for
 MIME messages on Internet.
 
-=head2 DEFINITIONS
+=head2 Definitions
 
 The B<charset> is ``character set'' used in MIME to refer to a
 method of converting a sequence of octets into a sequence of characters.
@@ -93,7 +93,7 @@ characters.
 =cut
 
 use strict;
-use vars qw(@ISA $VERSION @EXPORT @EXPORT_OK %EXPORT_TAGS);
+use vars qw(@ISA $VERSION @EXPORT @EXPORT_OK %EXPORT_TAGS $Config);
 use Exporter;
 @ISA = qw(Exporter);
 @EXPORT = qw(body_encoding canonical_charset header_encoding output_charset
@@ -110,7 +110,7 @@ use Carp qw(croak);
 use constant USE_ENCODE => ($] >= 5.008001)? 'Encode': '';
 
 my @ENCODE_SUBS = qw(FB_CROAK FB_PERLQQ FB_HTMLCREF FB_XMLCREF
-		     decode encode from_to is_utf8 resolve_alias);
+		     is_utf8 resolve_alias);
 if (USE_ENCODE) {
     eval "use ".USE_ENCODE." \@ENCODE_SUBS;";
 } else {
@@ -121,7 +121,7 @@ if (USE_ENCODE) {
     }
 }
 
-$VERSION = '1.000';
+$VERSION = '1.006.1';
 
 ######## Private Attributes ########
 
@@ -155,6 +155,9 @@ my %CHARSETS = (# input		    header enc body enc output conv
 		'ISO-2022-JP' =>	['B',	undef,	undef],
 		'KOI8-R' =>		['B',	'B',	undef],
 		'UTF-8' =>		['S',	'B',	undef],
+		'HZ-GB-2312' =>		['B',	undef,	undef],
+		'UTF-7' =>		['Q',	undef,	undef],
+		'GSM03.38' =>		[undef,	undef,	undef], # not for MIME
 		# We're making this one up to represent raw unencoded 8bit
 		'8BIT' =>		[undef,	'B',	'ISO-8859-1'],
 		);
@@ -172,13 +175,62 @@ my %CHARSET_ALIASES = (# unpreferred		preferred
 		       "KS_C_5601" =>		"KS_C_5601-1987",
 		       "SHIFTJIS" =>		"SHIFT_JIS",
 		       "SHIFTJISX0213" =>	"SHIFT_JISX0213",
-		       "UNICODE-1-1-UTF-7" =>	"UTF-7",
+		       "UNICODE-1-1-UTF-7" =>	"UTF-7", # RFC 1642 (obs.)
 		       "UTF8" =>		"UTF-8",
-		       "UTF-8-STRICT" =>	"UTF-8",
+		       "UTF-8-STRICT" =>	"UTF-8", # Perl internal use
+		       "HZ" =>			"HZ-GB-2312", # RFC 1842
+		       "GSM0338" =>		"GSM03.38",
 		       );
 
-# ISO-2022-* escape sequnces to detect charset from unencoded data.
-my @ISO2022_SEQ = (# escape seq	possible charset
+# Some vendors encode characters beyond standardized mappings using extended
+# encoders.  Some other standard encoders need additional encode modules.
+my %ENCODERS = (
+		'EXTENDED' => {
+		    'ISO-8859-1' => [['cp1252'], ],     # Encode::Byte
+		    'ISO-8859-2' => [['cp1250'], ],     # Encode::Byte
+		    'ISO-8859-5' => [['cp1251'], ],     # Encode::Byte
+		    'ISO-8859-6' => [
+				     ['cp1256'],        # Encode::Byte
+				     # ['cp1006'],      # ditto, for Farsi
+				    ],
+		    'ISO-8859-7' => [['cp1253'], ],     # Encode::Byte
+		    'ISO-8859-8' => [['cp1255'], ],     # Encode::Byte
+		    'ISO-8859-9' => [['cp1254'], ],     # Encode::Byte
+		    'ISO-8859-13'=> [['cp1257'], ],     # Encode::Byte
+		    'GB2312'     => [['cp936'], ],      # Encode::CN
+		    'EUC-JP'     => [['cp51932',        'Encode::EUCJPMS'], ],
+		    'ISO-2022-JP'=> [
+				     ['cp50220',        'Encode::EUCJPMS'],
+				     # ['cp50221',      'Encode::EUCJPMS'],
+				     ['iso-2022-jp-ms', 'Encode::ISO2022JPMS'],
+				     ['iso-2022-jp-1'], # Encode::JP (note*)
+				    ],
+		    'SHIFT_JIS'  => [['cp932'], ],      # Encode::JP
+		    'EUC-KR'     => [['cp949'], ],      # Encode::KR
+		    'BIG5'       => [
+				     # ['big5plus',     'Encode::HanExtra'],
+				     # ['big5-2003',    'Encode::HanExtra'], 
+				     ['cp950'],         # Encode::TW
+				     # ['big5-1984',    'Encode::HanExtra'], 
+				    ],
+		    'TIS620'     => [['iso-8859-11'], ], # Encode::Byte; NBSP
+		    'UTF-8'      => [['utf8'], ],       # Special name on Perl
+		},
+		'STANDARD' => {
+		    'GB18030'       => [['gb18030',     'Encode::HanExtra'], ],
+		    'EUC-JISX0213'  => [['euc-jisx0213', 'Encode::JIS2K'], ],
+		    'ISO-2022-JP-3' => [['iso-2022-jp-3', 'Encode::JIS2K'], ],
+		    'SHIFT_JISX0213'=> [['shiftjisx0213', 'Encode::JIS2K'], ],
+		    'EUC-TW'        => [['euc-tw',      'Encode::HanExtra'], ],
+		    'HZ-GB-2312'    => [['hz'], ],	# Encode::CN
+		    'GSM03.38'      => [['gsm0338'], ],	# Encode::GSM0338
+		},
+);
+
+# ISO-2022-* escape sequences etc. to detect charset from unencoded data.
+my @ESCAPE_SEQS = ( 
+		# ISO-2022-* sequences
+		   # escape seq, possible charset
 		   # Following sequences are commonly used.
 		   ["\033\$\@",	"ISO-2022-JP"],	# RFC 1468
 		   ["\033\$B",	"ISO-2022-JP"],	# ditto
@@ -193,39 +245,91 @@ my @ISO2022_SEQ = (# escape seq	possible charset
 		   ["\033\$*H",	"ISO-2022-CN"], # ditto
 		   # Other sequences will be used with appropriate charset
 		   # parameters, or hardly used.
-		   );
 
-		   # note*: This RFC defines ISO-2022-JP-1, superset of
+		   # note*: This RFC defines ISO-2022-JP-1, superset of 
 		   # ISO-2022-JP.  But that charset name is rarely used.
-		   # OTOH many of codecs for ISO-2022-JP recognize this
+		   # OTOH many of encoders for ISO-2022-JP recognize this
 		   # sequence so that comatibility with EUC-JP will be
 		   # guaranteed.
 
+		# Singlebyte 7-bit sequences
+		   # escape seq, possible charset
+		   ["\033e",	"GSM03.38"],	# ESTI GSM 03.38 (note*)
+		   ["\033\012",	"GSM03.38"],	# ditto
+		   ["\033<",	"GSM03.38"],	# ditto
+		   ["\033/",	"GSM03.38"],	# ditto
+		   ["\033>",	"GSM03.38"],	# ditto
+		   ["\033\024",	"GSM03.38"],	# ditto
+		   ["\033(",	"GSM03.38"],	# ditto
+		   ["\033\@",	"GSM03.38"],	# ditto
+		   ["\033)",	"GSM03.38"],	# ditto
+		   ["\033=",	"GSM03.38"],	# ditto
+
+		   # note*: This is not used for MIME message.
+		  );
+
+######## Public Configuration Attributes ########
+
+$Config = {
+    Detect7bit =>      'YES',
+    Mapping =>         'EXTENDED',
+    Replacement =>     'DEFAULT',
+};
+eval { require MIME::Charset::Defaults; };
+
 ######## Private Constants ########
 
-my $NONASCIIRE = qr{
+my $NON7BITRE = qr{
     [^\x01-\x7e]
 }x;
 
+my $NONASCIIRE = qr{
+    [^\x09\x0a\x0d\x20\x21-\x7e]
+}x;
+
 my $ISO2022RE = qr{
-    ^ISO-2022-
+    ISO-2022-.+
+}ix;
+
+my $ASCIITRANSRE = qr{
+    HZ-GB-2312 | UTF-7
 }ix;
 
 
 ######## Public Functions ########
 
-=head2 CONSTRUCTOR
+=head2 Constructor
 
-=item $charset = MIME::Charset->new(CHARSET)
+=item $charset = MIME::Charset->new([CHARSET [, OPTS]])
 
-Create charset object from CHARSET.
+Create charset object.
+
+OPTS may accept following key-value pairs.
+B<NOTE>:
+When Unicode/multibyte support is disabled (see L<"USE_ENCODE">),
+conversion will not be performed.  So these options do not have any effects.
+
+=over 4
+
+=item Mapping => MAPTYPE
+
+Specify extended mappings actually used for charset names.
+C<"EXTENDED"> uses extended mappings.
+C<"STANDARD"> uses standardized strict mappings.
+Default is C<"EXTENDED">.
+
+=back
 
 =cut
 
 sub new {
     my $class = shift;
     my $charset = shift;
-    return undef unless $charset;
+    return bless {}, $class unless $charset;
+    my %params = @_;
+    my $mapping = uc($params{'Mapping'} || $Config->{Mapping});
+
+    $charset = "HZ" if $charset =~ /\bhz.?gb.?2312$/i; # workaround
     $charset = resolve_alias($charset) || $charset;
     $charset = $CHARSET_ALIASES{uc($charset)} || uc($charset);
     my ($henc, $benc, $outcset);
@@ -236,16 +340,51 @@ sub new {
     } else {
 	($henc, $benc, $outcset) = ('S', 'B', undef);
     }
-    
+    my ($decoder, $encoder);
+    if (USE_ENCODE) {
+	$decoder = _find_encoder($charset, $mapping);
+	$encoder = _find_encoder($outcset, $mapping);
+    } else {
+	$decoder = $encoder = undef;
+    }
+
     bless {
 	InputCharset => $charset,
+	Decoder => $decoder,
 	HeaderEncoding => $henc,
 	BodyEncoding => $benc,
 	OutputCharset => ($outcset || $charset),
+	Encoder => ($encoder || $decoder),
     }, $class;
 }
 
-=head2 GETTING INFORMATIONS OF CHARSETS
+sub _find_encoder($$) {
+    my $charset = uc(shift);
+    return undef unless $charset;
+    my $mapping = uc(shift);
+    my ($spec, $name, $module, $encoder);
+
+    my $preserveerr = $@;
+    foreach my $m (('EXTENDED', 'STANDARD')) {
+	next if $m eq 'EXTENDED' and $mapping ne 'EXTENDED';
+	$spec = $ENCODERS{$m}->{$charset};
+	next unless $spec;
+	foreach my $s (@{$spec}) {
+	    ($name, $module) = @{$s};
+	    if ($module) {
+		eval "use $module;";
+		next if $@;
+	    }
+	    $@ = $preserveerr;
+	    $encoder = Encode::find_encoding($name);
+	    return $encoder if ref $encoder;
+	}
+    }
+    $@ = $preserveerr;
+    return Encode::find_encoding($charset);
+}
+
+=head2 Getting Informations of Charsets
 
 =item $charset->body_encoding
 
@@ -262,25 +401,79 @@ not be same as encoding for message header.
 sub body_encoding($) {
     my $self = shift;
     return undef unless $self;
-    $self = MIME::Charset->new($self) unless ref $self;
+    $self = __PACKAGE__->new($self) unless ref $self;
     $self->{BodyEncoding};
 }
 
-=item $charset->canonical_charset
+=item $charset->as_string
 
 =item canonical_charset CHARSET
 
-Get canonical name for charset CHARSET.
+Get canonical name for charset.
 
 =cut
 
 sub canonical_charset($) {
     my $self = shift;
     return undef unless $self;
-    $self = MIME::Charset->new($self) unless ref $self;
+    $self = __PACKAGE__->new($self) unless ref $self;
     $self->{InputCharset};
 }
 
+sub as_string($) {
+    my $self = shift;
+    $self->{InputCharset};
+}
+
+=item $charset->decoder
+
+Get L<"Encode::Encoding"> object to decode strings by charset.
+
+=cut
+
+sub decoder($) {
+    my $self = shift;
+    $self->{Decoder};
+}
+
+=item $charset->dup
+
+Get a copy of charset object.
+
+=cut
+
+sub dup($) {
+    my $self = shift;
+    my $obj = __PACKAGE__->new(undef);
+    %{$obj} = %{$self};
+    $obj;
+}
+
+=item $charset->encoder([CHARSET])
+
+Get L<"Encode::Encoding"> object to encode Unicode string using compatible
+charset recommended to be used for messages on Internet.
+
+If optional CHARSET is specified, replace encoder (and output charset
+name) of $charset object with those of CHARSET, therefore,
+$charset object will be a converter between original charset and
+new CHARSET.
+
+=cut
+
+sub encoder($$;) {
+    my $self = shift;
+    my $charset = shift;
+    if ($charset) {
+	$charset = __PACKAGE__->new($charset) unless ref $charset;
+	$self->{OutputCharset} = $charset->{InputCharset};
+	$self->{Encoder} = $charset->{Decoder};
+	$self->{BodyEncoding} = $charset->{BodyEncoding};
+	$self->{HeaderEncoding} = $charset->{HeaderEncoding};
+    }
+    $self->{Encoder};
+}
+
 =item $charset->header_encoding
 
 =item header_encoding CHARSET
@@ -296,7 +489,7 @@ for message body.
 sub header_encoding($) {
     my $self = shift;
     return undef unless $self;
-    $self = MIME::Charset->new($self) unless ref $self;
+    $self = __PACKAGE__->new($self) unless ref $self;
     $self->{HeaderEncoding};
 }
 
@@ -316,11 +509,11 @@ return the result of L<"canonical_charset">.
 sub output_charset($) {
     my $self = shift;
     return undef unless $self;
-    $self = MIME::Charset->new($self) unless ref $self;
+    $self = __PACKAGE__->new($self) unless ref $self;
     $self->{OutputCharset};
 }
 
-=head2 TRANSLATING TEXT DATA
+=head2 Translating Text Data
 
 =item $charset->body_encode(STRING [, OPTS])
 
@@ -337,15 +530,15 @@ conversion will not be performed.  So these options do not have any effects.
 
 =over 4
 
-=item Replacement => REPLACEMENT
-
-Specifies error handling scheme.  See L<"ERROR HANDLING">.
-
 =item Detect7bit => YESNO
 
 Try auto-detecting 7-bit charset when CHARSET is not given.
 Default is C<"YES">.
 
+=item Replacement => REPLACEMENT
+
+Specifies error handling scheme.  See L<"Error Handling">.
+
 =back
 
 3-item list of (I<converted string>, I<charset for output>,
@@ -366,28 +559,20 @@ sub body_encode {
 	$text = shift;
     } else {
 	$text = $self;
-	$self = MIME::Charset->new(shift);
+	$self = __PACKAGE__->new(shift);
     }
-    my ($encoded, $charset, $cset) = &_text_encode($self, $text, @_);
+    my ($encoded, $charset) = $self->_text_encode($text, @_);
+    return ($encoded, undef, 'BASE64')
+	unless $charset and $charset->{InputCharset};
+    my $cset = $charset->{OutputCharset};
 
     # Determine transfer-encoding.
-    my $enc;
-    my $dummy = $encoded;
-    eval {
-	from_to($dummy, $cset, "US-ASCII", FB_CROAK());
-    };
-    if (!$@ and $dummy eq $encoded) {
-	$cset = "US-ASCII";
-	$enc = undef;
-    } else {
-	$@ = '';
-	$enc = $charset->body_encoding;
-    }
+    my $enc = $charset->{BodyEncoding};
 
     if (!$enc and $encoded !~ /\x00/) {	# Eliminate hostile NUL character.
-        if ($encoded =~ $NONASCIIRE) {	# String contains 8bit char(s).
+        if ($encoded =~ $NON7BITRE) {	# String contains 8bit char(s).
             $enc = '8BIT';
-	} elsif ($cset =~ $ISO2022RE) {	# ISO-2022-* outputs are 7BIT.
+	} elsif ($cset =~ /^($ISO2022RE|$ASCIITRANSRE)$/) {	# 7BIT.
             $enc = '7BIT';
         } else {			# Pure ASCII.
             $enc = '7BIT';
@@ -403,6 +588,47 @@ sub body_encode {
     return ($encoded, $cset, $enc);
 }
 
+=item $charset->decode(STRING [,CHECK])
+
+Decode STRING to Unicode.
+
+B<Note>:
+When Unicode/multibyte support is disabled (see L<"USE_ENCODE">),
+this function will die.
+
+=cut
+
+sub decode($$$;) {
+    my $self = shift;
+    my $s = shift;
+    my $check = shift || 0;
+    $self->{Decoder}->decode($s, $check);
+}
+
+=item $charset->encode(STRING [,CHECK])
+
+Encode STRING (Unicode or non-Unicode) using compatible charset recommended
+to be used for messages on Internet (if this module knows it).
+Note that string will be decoded then encoded even if compatible charset
+was equal to original charset.
+
+B<Note>:
+When Unicode/multibyte support is disabled (see L<"USE_ENCODE">),
+this function will die.
+
+=cut
+
+sub encode($$$;) {
+    my $self = shift;
+    my $s = shift;
+    my $check = shift || 0;
+
+    unless (is_utf8($s) or $s =~ /[^\x00-\xFF]/) {
+	$s = $self->{Decoder}->decode($s, ($check & 0x1)? FB_CROAK(): 0);
+    }
+    $self->{Encoder}->encode($s, $check);
+}
+
 =item $charset->encoded_header_len(STRING [, ENCODING])
 
 =item encoded_header_len STRING, ENCODING, CHARSET
@@ -425,7 +651,7 @@ sub encoded_header_len($$$;) {
 	$s = $self;
 	$encoding = uc(shift);
 	$self  = shift;
-	$self = MIME::Charset->new($self) unless ref $self;
+	$self = __PACKAGE__->new($self) unless ref $self;
     }
 
     #FIXME:$encoding === undef
@@ -449,8 +675,7 @@ sub _enclen_B($) {
 
 sub _enclen_Q($) {
     my $s = shift;
-    my @o;
-    @o = ($s =~ /(\?|=|_|[^ \x21-\x7e])/gos);
+    my @o = ($s =~ m{([^- !*+/0-9A-Za-z])}gos);
     length($s) + scalar(@o) * 2;
 }
 
@@ -469,15 +694,15 @@ conversion will not be performed.  So these options do not have any effects.
 
 =over 4
 
-=item Replacement => REPLACEMENT
-
-Specifies error handling scheme.  See L<"ERROR HANDLING">.
-
 =item Detect7bit => YESNO
 
 Try auto-detecting 7-bit charset when CHARSET is not given.
 Default is C<"YES">.
 
+=item Replacement => REPLACEMENT
+
+Specifies error handling scheme.  See L<"Error Handling">.
+
 =back
 
 3-item list of (I<converted string>, I<charset for output>,
@@ -501,27 +726,18 @@ sub header_encode {
 	$text = shift;
     } else {
 	$text = $self;
-	$self = MIME::Charset->new(shift);
+	$self = __PACKAGE__->new(shift);
     }
-    my ($encoded, $charset, $cset) = &_text_encode($self, $text, @_);
-    return ($encoded, '8BIT', undef) unless $cset;
+    my ($encoded, $charset) = $self->_text_encode($text, @_);
+    return ($encoded, '8BIT', undef)
+	unless $charset and $charset->{InputCharset};
+    my $cset = $charset->{OutputCharset};
 
     # Determine encoding scheme.
-    my $enc;
-    my $dummy = $encoded;
-    eval {
-	from_to($dummy, $cset, "US-ASCII", FB_CROAK());
-    };
-    if (!$@ and $dummy eq $encoded) {
-	$cset = "US-ASCII";
-	$enc = undef;
-    } else {
-	$@ = '';
-	$enc = $charset->{HeaderEncoding};
-    }
+    my $enc = $charset->{HeaderEncoding};
 
-    if (!$enc and $encoded !~ $NONASCIIRE) {
-	unless ($cset =~ $ISO2022RE) {	# ISO-2022-* outputs are 7BIT.
+    if (!$enc and $encoded !~ $NON7BITRE) {
+	unless ($cset =~ /^($ISO2022RE|$ASCIITRANSRE)$/) {	# 7BIT.
             $cset = 'US-ASCII';
         }
     } elsif ($enc eq 'S') {
@@ -540,83 +756,99 @@ sub _text_encode {
     my $charset = shift;
     my $s = shift;
     my %params = @_;
-    my $replacement = uc($params{'Replacement'}) || "DEFAULT";
-    my $detect7bit = uc($params{'Detect7bit'}) || "YES";
+    my $replacement = uc($params{'Replacement'} || $Config->{Replacement});
+    my $detect7bit = uc($params{'Detect7bit'} || $Config->{Detect7bit});
+    my $encoding = $params{'Encoding'} ||
+	(exists $params{'Encoding'}? undef: 'A'); # undocumented
 
-    if (!$charset) {
-	if ($s =~ $NONASCIIRE) {
-	    return ($s, undef, undef);
+    if (!$encoding or $encoding ne 'A') { # no 7-bit auto-detection
+	$detect7bit = 'NO';
+    }
+    unless ($charset->{InputCharset}) {
+	if ($s =~ $NON7BITRE) {
+	    return ($s, undef);
 	} elsif ($detect7bit ne "NO") {
-	    $charset = MIME::Charset->new(&_detect_7bit_charset($s));
+	    $charset = __PACKAGE__->new(&_detect_7bit_charset($s));
 	} else {
-	    $charset = MIME::Charset->new($DEFAULT_CHARSET);
+	    $charset = __PACKAGE__->new($DEFAULT_CHARSET,
+					Mapping => 'STANDARD');
 	} 
     }
-
-    # Unknown charset.
-    return ($s, $charset, $charset->{InputCharset})
-	unless resolve_alias($charset->{InputCharset});
+    if (!$encoding or $encoding ne 'A') { # no conversion
+	$charset = $charset->dup;
+	$charset->encoder($charset);
+	$charset->{HeaderEncoding} = $encoding;
+	$charset->{BodyEncoding} = $encoding;
+    }
+    my $check = ($replacement and $replacement =~ /^\d+$/)?
+	$replacement:
+    {
+	'CROAK' => FB_CROAK(),
+	'STRICT' => FB_CROAK(),
+	'FALLBACK' => FB_CROAK(), # special
+	'PERLQQ' => FB_PERLQQ(),
+	'HTMLCREF' => FB_HTMLCREF(),
+	'XMLCREF' => FB_XMLCREF(),
+    }->{$replacement || ""} || 0;
 
     # Encode data by output charset if required.  If failed, fallback to
     # fallback charset.
-    my $cset = $charset->output_charset;
     my $encoded;
-
-    if (is_utf8($s) or $s =~ /[^\x00-\xFF]/) {
-	if ($replacement =~ /^(?:CROAK|STRICT|FALLBACK)$/) {
+    if (is_utf8($s) or $s =~ /[^\x00-\xFF]/ or
+	($charset->{InputCharset} || "") ne ($charset->{OutputCharset} || "")) {
+	if ($check & 0x1) { # CROAK or FALLBACK
 	    eval {
 		$encoded = $s;
-		$encoded = encode($cset, $encoded, FB_CROAK());
+		$encoded = $charset->encode($encoded, FB_CROAK());
 	    };
 	    if ($@) {
 		if ($replacement eq "FALLBACK" and $FALLBACK_CHARSET) {
-		    $cset = $FALLBACK_CHARSET;
+		    my $cset = __PACKAGE__->new($FALLBACK_CHARSET,
+						Mapping => 'STANDARD');
+		    # croak unknown charset
+		    croak "unknown charset ``$FALLBACK_CHARSET''"
+			unless $charset->{Decoder};
+		    # charset transformation
+		    $charset->encoder($cset);
 		    $encoded = $s;
-		    $encoded = encode($cset, $encoded);
-		    $charset = MIME::Charset->new($cset);
+		    $encoded = $charset->encode($encoded, 0);
+		    $cset->encoder($cset);
+		    $charset = $cset;
 		} else {
+		    $@ =~ s/ at .+$//;
 		    croak $@;
 		}
 	    }
-	} elsif ($replacement eq "PERLQQ") {
-	    $encoded = encode($cset, $s, FB_PERLQQ());
-	} elsif ($replacement eq "HTMLCREF") {
-	    $encoded = encode($cset, $s, FB_HTMLCREF());
-	} elsif ($replacement eq "XMLCREF") {
-	    $encoded = encode($cset, $s, FB_XMLCREF());
 	} else {
-	    $encoded = encode($cset, $s);
+	    $encoded = $s;
+	    $encoded = $charset->encode($encoded, $check);
 	}
-    } elsif ($charset->{InputCharset} ne $cset) {
-	$encoded = $s;
-	if ($replacement =~ /^(?:CROAK|STRICT|FALLBACK)$/) {
-	    eval {
-		from_to($encoded, $charset->{InputCharset}, $cset, FB_CROAK());
-	    };
-	    if ($@) {
-		if ($replacement eq "FALLBACK" and $FALLBACK_CHARSET) {
-		    $cset = $FALLBACK_CHARSET;
-		    $encoded = $s;
-		    from_to($encoded, $charset->{InputCharset}, $cset);
-		    $charset = MIME::Charset->new($cset);
-		} else {
-		    croak $@;
-		}
-	    }
-        } elsif ($replacement eq "PERLQQ") {
-            from_to($encoded, $charset->{InputCharset}, $cset, FB_PERLQQ());
-        } elsif ($replacement eq "HTMLCREF") {
-            from_to($encoded, $charset->{InputCharset}, $cset, FB_HTMLCREF());
-        } elsif ($replacement eq "XMLCREF") {
-            from_to($encoded, $charset->{InputCharset}, $cset, FB_XMLCREF());
-        } else {
-            from_to($encoded, $charset->{InputCharset}, $cset);
-        }
     } else {
         $encoded = $s;
     }
 
-    return ($encoded, $charset, $cset);
+    if ($encoded !~ /$NONASCIIRE/) { # maybe ASCII
+	# check ``ASCII transformation'' charsets
+	if ($charset->{OutputCharset} =~ /^($ASCIITRANSRE)$/ and
+	    $encoded =~ /[+~]/) {
+	    my $u = $encoded;
+	    if ($charset->encoder) {
+		$u = $charset->encoder->decode($encoded); # dec. by output
+	    } elsif (!USE_ENCODE) { # workaround for pre-Encode environment
+		$u = "x$u";
+	    } else { # NOTREACHED
+		croak __PACKAGE__.": bug in _text_encode.  Report developer.";
+	    }
+	    $charset->encoder(__PACKAGE__->new($DEFAULT_CHARSET,
+					       Mapping => 'STANDARD'))
+		if $u eq $encoded;
+	} elsif ($charset->{OutputCharset} ne "US-ASCII") {
+	    $charset->encoder(__PACKAGE__->new($DEFAULT_CHARSET,
+					       Mapping => 'STANDARD'));
+	}
+    }
+
+    return ($encoded, $charset);
 }
 
 sub _detect_7bit_charset($) {
@@ -624,27 +856,47 @@ sub _detect_7bit_charset($) {
     my $s = shift;
     return $DEFAULT_CHARSET unless $s;
 
-    # Try to detect ISO-2022-* escape sequences.
-    foreach (@ISO2022_SEQ) {
+    # Try to detect 7-bit escape sequences.
+    foreach (@ESCAPE_SEQS) {
 	my ($seq, $cset) = @$_;
 	if (index($s, $seq) >= 0) {
+            my $decoder = __PACKAGE__->new($cset);
+            next unless $decoder and $decoder->{Decoder};
             eval {
 		my $dummy = $s;
-		decode($cset, $dummy, FB_CROAK());
+		$decoder->decode($dummy, FB_CROAK());
 	    };
 	    if ($@) {
 		next;
 	    }
-	    return $cset;
+	    return $decoder->{InputCharset};
 	}
     }
 
-    # How about HZ, VIQR, ...?
+    # How about HZ, VIQR, UTF-7, ...?
 
     return $DEFAULT_CHARSET;
 }
 
-=head2 MANIPULATING MODULE DEFAULTS
+=item $charset->undecode(STRING [,CHECK])
+
+Encode Unicode string STRING to byte string by input charset of $charset.
+This is equivalent to C<$charset-E<gt>decoder-E<gt>encode()>.
+
+B<Note>:
+When Unicode/multibyte support is disabled (see L<"USE_ENCODE">),
+this function will die.
+
+=cut
+
+sub undecode($$$;) {
+    my $self = shift;
+    my $s = shift;
+    my $check = shift || 0;
+    $self->{Decoder}->encode($s, $check);
+}
+
+=head2 Manipulating Module Defaults
 
 =over 4
 
@@ -774,22 +1026,22 @@ sub recommended ($;$;$;$) {
 	$CHARSETS{$charset} = \@spec;
 	return @spec;
     } else {
-	$charset = MIME::Charset->new($charset) unless ref $charset;
+	$charset = __PACKAGE__->new($charset) unless ref $charset;
 	return map { $charset->{$_} } qw(HeaderEncoding BodyEncoding
 					 OutputCharset);
     }
 }
 
-=head2 CONSTANTS
+=head2 Constants
 
 =item USE_ENCODE
 
 Unicode/multibyte support flag.
-Non-null string will be set when Unicode and multibyte support is enabled.
-Currently, this flag will be non-null on Perl 5.8.1 or later and
-null string on earlier versions of Perl.
+Non-empty string will be set when Unicode and multibyte support is enabled.
+Currently, this flag will be non-empty on Perl 5.8.1 or later and
+empty string on earlier versions of Perl.
 
-=head2 ERROR HANDLING
+=head2 Error Handling
 
 L<"body_encode"> and L<"header_encode"> accept following C<Replacement>
 options:
@@ -821,11 +1073,22 @@ Synonym is C<"STRICT">.
 Use C<FB_PERLQQ>, C<FB_HTMLCREF> or C<FB_XMLCREF>
 scheme defined by L<Encode> module.
 
+=item numeric values
+
+Numeric values are also allowed.
+For more details see L<Encode/Handling Malformed Data>.
+
 =back
 
 If error handling scheme is not specified or unknown scheme is specified,
 C<"DEFAULT"> will be assumed.
 
+=head2 Configuration File
+
+Built-in defaults for option parameters can be overridden by configuration
+file: F<MIME/Charset/Defaults.pm>.
+For more details read F<MIME/Charset/Defaults.pm.sample>.
+
 =head1 VERSION
 
 Consult $VERSION variable.
